/*
 * The MIT License
 *
 * Copyright 2011 Jacob Dilles <jdilles@gmu.edu>.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package edu.gmu.ece445.lc2200;

import java.io.PrintStream;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Holds more info than the Encoded Instruction, and provides an abstract 
 * method for encoding 
 * 
 * @author Jacob Dilles <jdilles@gmu.edu>
 */
public class AssemblyInstruction extends EncodedInstruction {

    private static final boolean DEBUG_PARSING = false;
    /** Deliminates arguments for display **/
    private static final String ARG_DELIM = ",  ";
    /** Magic number for unlinked branch targets **/
    public static final int BTMAGIC = 912090;
    /** Holds the 32 bit memory address of the instruction **/
    protected int address;
    /** Holds the address label for the instruction **/
    protected String label;
    /** Holds the comment string  */
    protected String comment;
    /** for branch instructions, the target address might change **/
    protected String branchTarget;

    public AssemblyInstruction(OpCode code, Register rx, Register ry, Register rz) {
        super();
        if (code != null) {
            setOpCode(code);
        }
        if (rx != null) {
            setX(rx);
        }
        if (ry != null) {
            setY(ry);
        }
        if (rz != null) {
            setZ(rz);
        }
    }

    public AssemblyInstruction(OpCode code, Register rx, Register ry) {
        super();
        if (code != null) {
            setOpCode(code);
        }
        if (rx != null) {
            setX(rx);
        }
        if (ry != null) {
            setY(ry);
        }
    }

    public AssemblyInstruction(OpCode code, Register rx, Register ry, String target) {
        super();
        if (code != null) {
            setOpCode(code);
        }
        if (rx != null) {
            setX(rx);
        }
        if (ry != null) {
            setY(ry);
        }
        this.branchTarget = target;
    }
    
    
    public AssemblyInstruction(OpCode code, Register rx, Register ry, int i) {
        super();
        if (code != null) {
            setOpCode(code);
        }
        if (rx != null) {
            setX(rx);
        }
        if (ry != null) {
            setY(ry);
        }
        setI(i);
    }

    public AssemblyInstruction(OpCode code) {
        super();
        if (code != null) {
            setOpCode(code);
        }

    }

    /**
     * Create a new instruction 
     * @param label
     * @param address
     * @param cmd
     * @param args
     * @param comment 
     */
    public AssemblyInstruction(String label, String address, String cmd, String args, String comment) {
        super(0);
        setLabel(label);
        if (address != null) {
            setAddress(Integer.decode(address));
        }
        setOpCode(cmd);
        parseArgsString(args);
        setComment(comment);
    }

    /**
     * Empty instruction 
     * @param word 
     */
    public AssemblyInstruction(int word) {
        super(0);
    }

    /**
     * Empty constructor 
     */
    public AssemblyInstruction() {
    }

    /**
     * Perform any additional steps required to encode the instruction 
     * @return 1 if linked, 0 otherwise 
     */
    public int encode(AssemblyProgram program, PrintStream out) {
        OpCode o = getInstruction();
        if ((o == OpCode.bne || o == OpCode.beq) && branchTarget != null) {
            AssemblyInstruction target = program.getInstructionByLabel(branchTarget);
            if (target == null) {
                throw new RuntimeException("Assembly Error: No instruction labeled " + branchTarget);
            }
            // Offset from PC+1
            int iValue = target.getAddress() - (address + 1);
            setI(iValue);
            out.printf(";### linked %s: %04d -> %04d (%d) \n", branchTarget, getAddress(), target.getAddress(), iValue);
            if (comment == null) {
                comment = String.format("(%d)", iValue);
            }
            return 1;
        }
        return 0;
    }

    /**
     * Set the branch target. This only makes sense for bne and beq
     * @param branchTarget 
     */
    public void setBranchTarget(AssemblyInstruction branchTarget) {
        this.branchTarget = branchTarget.getLabel();
    }

    /**
     * Get the branch target. This only makes sense for bne and beq
     * @return 
     */
    public String getBranchTargetLabel() {
        return branchTarget;
    }

    /**
     * Get the program address
     * @return 
     */
    public int getAddress() {
        return address;
    }

    /**
     * Set the program address
     * @param address 
     */
    public void setAddress(int address) {
        this.address = address;
    }

    /**
     * Get the comment string
     * @return 
     */
    public String getComment() {
        return comment;
    }

    /**
     * Set the comment string (null clears)
     * @param comment 
     */
    public void setComment(String comment) {
        this.comment = comment;
    }

    /**
     * Get the address label 
     * @return label, or null if no label 
     */
    public String getLabel() {
        return label;
    }

    /**
     * Set the address label (null clears)
     * @param label 
     */
    public void setLabel(String label) {
        this.label = label;
    }

    /**
     * Get the instruction name
     * @return 
     */
    public String getName() {
        return getInstruction().name();
    }

    @Override
    public String toString() {
        // max width is "$zero,  1048575($zero)" = 22 chars (never gonna happen)
//        return String.format("%10s %04x  %-5s %-25s %s", (label == null ? "" : label + ":"), address, getName(), getArgsString(), (comment == null || comment.isEmpty() ? "" : ";" + comment));
        return String.format("%10s %04d  %-5s %-25s %s", (label == null ? "" : label + ":"), address, getName(), getArgsString(), (comment == null || comment.isEmpty() ? "" : ";" + comment));
    }

    /**
     * Get the args string - automatically picks the right one for HW instructions
     * @return 
     */
    public String getArgsString() {
        switch (getInstruction()) {
            // O type inst
            case halt:
            case nop:
                return getArgsStringO();
            // J type 
            case jalr:
                return getArgsStringJ();
            // R type insts
            case add:
            case nand:
            case nor:
            case sub:
                return getArgsStringR();
            // I type insts
            case addi:
            case sll:
            case srl:
            case sra:
                return getArgsStringI();
            // I type, lw and sw are special
            case lw:
            case sw:
                return getArgsStringW();
            case beq:
            case bne:
                return getArgsStringB();
        }
        return "Unknown opcode: " + getOpCode();
    }

    /**
     * Get O type args string
     * @return 
     */
    protected String getArgsStringO() {
        return "";
    }

    /**
     * Get J type args string
     * @return 
     */
    protected String getArgsStringJ() {
        return getRegisterX() + ARG_DELIM + getRegisterY();
    }

    /**
     * Get R type args string
     * @return 
     */
    protected String getArgsStringR() {
        return getRegisterX() + ARG_DELIM + getRegisterY() + ARG_DELIM + getRegisterZ();
    }

    /**
     * Get I type args string
     * @return 
     */
    protected String getArgsStringI() {
        return getRegisterX() + ARG_DELIM + getRegisterY() + ARG_DELIM + getI();
    }

    /**
     * Get W type args string
     * @return 
     */
    protected String getArgsStringW() {
        return getRegisterX() + ARG_DELIM + getI() + "(" + getRegisterY() + ")";
    }

    /**
     * Get B type args string
     * @return 
     */
    protected String getArgsStringB() {
        String bl = getBranchTargetLabel();
        return getRegisterX() + ARG_DELIM + getRegisterY() + ARG_DELIM + (bl == null ? getI() : bl);
    }

    /**
     * Set the op code based on the string alias
     * @param opCodeAlias 
     */
    public void setOpCode(String opCodeAlias) {
        try {
            setOpCode(OpCode.valueOfName(opCodeAlias));
        } catch (IllegalArgumentException ex) {
            throw new IllegalArgumentException("No such instruction: " + opCodeAlias, ex);
        }
    }

    /**
     * Parses args to an instruction line 
     * @param args 
     */
    protected void parseArgsString(String args) {
        try {
            switch (getInstruction()) {
                // J type 
                case jalr:
                    parseJargs(args);
                    return;
                // R type insts
                case add:
                case nand:
                case nor:
                case sub:
                    parseTypeRargs(args);
                    return;
                // I type insts
                case addi:
                case sll:
                case srl:
                case sra:
                    parseTypeIargs(args);
                    return;
                // I type, lw and sw are special
                case lw:
                case sw:
                    parseWargs(args);
                    return;
                case beq:
                case bne:
                    parseBranchArgs(args);
                    return;
                // O type inst
                case nop:
                case halt:
                default:
                    return;
            }
        } catch (Exception ex) {
            throw new IllegalArgumentException("Error parsing instruction argument \"" + args + "\"", ex);
        }
    }

    /**
     * Parses arguments for J type instructions 
     * @param args 
     */
    protected void parseJargs(String args) {
        String[] parts = args.split(",");
        if (parts.length != 2) {
            throw new IllegalArgumentException("JALR type should have 2 args: " + args);
        }
        Register rx = Register.valueOfName(parts[0].trim());
        Register ry = Register.valueOfName(parts[1].trim());
        setX(rx);
        setY(ry);
    }

    /**
     * Parses arguments for LW/SW instructions
     * @param args 
     */
    protected void parseWargs(String args) {
        if (DEBUG_PARSING) {
            System.out.println("LW/SW args=" + args);
        }
        String[] parts = args.split(",");
        if (parts.length != 2) {
            throw new IllegalArgumentException("LW/SW type should have 2 args: " + args);
        }
        Register rx = Register.valueOfName(parts[0].trim());
        setX(rx);
        if (DEBUG_PARSING) {
            System.out.println("LW/SW x=" + rx);
        }

        parts = parts[1].trim().split("\\(");
        if (parts.length != 2) {
            throw new IllegalArgumentException("LW/SW syntax: " + args);
        }

        int i = Integer.decode(parts[0]);
        setI(i);
        if (DEBUG_PARSING) {
            System.out.println("LW/SW i=" + i);
        }


        parts = parts[1].trim().split("\\)");
//        if (parts.length ) {
//            throw new IllegalArgumentException("LW/SW syntax: " + args);
//        }
        Register ry = Register.valueOfName(parts[0].trim());
        setY(ry);

        if (DEBUG_PARSING) {
            System.out.println("LW/SW ry=" + ry);
        }
    }

    /**
     * Parses branch instruction arguments 
     * @param args 
     */
    protected void parseBranchArgs(String args) {
        String[] parts = args.split(",");
        if (parts.length != 3) {
            throw new IllegalArgumentException("I type should have 3 args");
        }
        Register rx = Register.valueOfName(parts[0].trim());
        Register ry = Register.valueOfName(parts[1].trim());
        String bt = parts[2].trim();
        try {
            int i = Integer.decode(bt);
            setI(i);
        } catch (Exception ex) {
            branchTarget = parts[2].trim();
            setI(BTMAGIC);
        }
        setX(rx);
        setY(ry);

    }

    /**
     * Parses I type args
     * @param args 
     */
    protected void parseTypeIargs(String args) {
        String[] parts = args.split(",");
        if (parts.length != 3) {
            throw new IllegalArgumentException("I type should have 3 args");
        }

        Register rx = Register.valueOfName(parts[0].trim());
        Register ry = Register.valueOfName(parts[1].trim());

        args = parts[2].replaceAll("[\\s]+", "");
        int i = Integer.decode(args);

        setX(rx);
        setY(ry);
        setI(i);
    }

    /**
     * Parse the R type args
     * @param args 
     */
    protected void parseTypeRargs(String args) {
        String[] parts = args.split(",");
        if (parts.length != 3) {
            throw new IllegalArgumentException("R type should have 3 args");
        }

        Register rx = Register.valueOfName(parts[0]);
        Register ry = Register.valueOfName(parts[1]);
        Register rz = Register.valueOfName(parts[2]);

        setX(rx);
        setY(ry);
        setZ(rz);
    }
}
