/*
 * The MIT License
 *
 * Copyright 2011 Jacob Dilles.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package edu.gmu.ece445.lc2200;

import java.io.Reader;
import java.util.List;
import java.util.Date;
import edu.gmu.ece445.lc2200.emulator.LC2200Impl;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.PrintStream;
import static edu.gmu.ece445.lc2200.Utils.*;

/**
 * A basic emulator 
 * @author Jacob Dilles
 */
public class Emulator {

    private LC2200 cpu;
    private boolean verbose = true;

    public Emulator(AssemblyProgram program) {
        cpu = new LC2200Impl(program.getEncodedProgram());
    }

    public void emulate(String args) {
        System.out.println("Welcome! Your program has assembled without errors.");
        System.out.println("\n\n-------------   LC-2200 Emulator ----------------------");
        System.out.println("The last instruction address is " + (cpu.getProgram().size() - 1));


        if (args.contains("p")) {
            cpu.setPc(getInteger("Initial PC (Program Counter) value", "0"));
        }

        if (args.contains("m")) {
            String memimage = getInput("Memory image file to load [enter for none]", null);
            if (memimage != null) {
                try {
                    Reader r = new FileReader(memimage);
                    cpu.getMemory().preload(r);
                    r.close();
                } catch (Exception ex) {
                    System.out.println("Error reading file... Clearning memory");
                }
            }
        }
        if (args.contains("q")) {
            verbose = false;
        }

        if (args.contains("r")) {
            try {
                run(-1);
            } catch (Exception ex) {
                System.out.println("Terminated: " + ex.getMessage());
            }
            cpu.clock();
            System.out.println(cpu.report());
            String dump = "dump" + System.currentTimeMillis() + ".txt";
            System.out.println("Dumping output to " + dump + "  - Please Wait...");
            dumpall(dump);
        } else {
            helpMenu();
            processInput();
        }
        System.out.println("Thank you for using the LC-2200 Emulator. Have a nice day!");
    }

    int run(int runto) {
        boolean running;

        do {
            running = cpu.clock();
            if (verbose) {
                System.out.println(cpu.report());
            }

            // autorun
            if (--runto == 0) {
                return 0;
            }
        } while (running);

        return runto;
    }

    void processInput() {
        String in = "";
        while (true) {
            try {
                in = getInput("LC2200>", null);

                // Pressed enter
                if (in == null || in.isEmpty()) {
                    run(1);
                    continue;
                }

                in = in.trim();

                // Run commands
                if (in.toLowerCase().startsWith("run")) {
                    in = in.trim();
                    if (in.length() == 3) {
                        run(-1);
                    } else {
                        run(Integer.decode(in.substring(4)));
                    }
                    continue;
                }

                // Set commands
                if (in.contains("=")) {

                    String[] parts = in.split("=");
                    if (parts.length != 2) {
                        System.out.println("Bad input: " + in);
                        continue;
                    }
                    int x = Integer.decode(parts[1]);
                    if (parts[0].equalsIgnoreCase("in")) {
                        cpu.getMemory().storeWord(MainMemory.INPUT_ADDRESS, x);
                        System.out.println("IN <= " + cpu.getMemory().loadWord(MainMemory.INPUT_ADDRESS));
                    } else if (parts[0].equalsIgnoreCase("pc")) {
                        cpu.setPc(x);
                        System.out.println("PC <= " + cpu.getPc());
                    } else {
                        try {
                            Register r = Register.valueOfName(parts[0]);
                            cpu.getRegisters().set(r, x);
                            System.out.println(r + " <= " + cpu.getRegisters().get(r));
                        } catch (IllegalArgumentException ex) {
                            int y = Integer.decode(parts[0]);
                            cpu.getMemory().storeWord(y, x);
                            System.out.println("M[" + y + "] <= " + cpu.getMemory().loadWord(y));
                        }
                    }
                    continue;
                }

                // Dump commands
                if (in.startsWith("rf")) {
                    dumpRegs(in.substring(2));
                    continue;
                }
                if (in.startsWith("md")) {
                    dumpMem(in.substring(2));
                    continue;
                }
                if (in.startsWith("ph")) {
                    dumpPH(in.substring(2));
                    continue;
                }
                if (in.startsWith("dump")) {
                    dumpall(null);
                    continue;
                }

                if (in.startsWith("reset")) {
                    cpu = new LC2200Impl(cpu.getProgram());
                    continue;
                }

                if (in.startsWith("q")) {
                    if (in.startsWith("quit")) {
                        if (in.startsWith("quitd")) {
                            dumpall(null);
                        }
                        return;
                    }
                    in = getInput("Are you sure you want to quit [y/N]", "n");
                    if (in == null || !in.toLowerCase().startsWith("y")) {
                        continue;
                    }
                    return;
                }

                helpMenu();
                continue;
            } catch (Exception ex) {
                System.out.println("Bad input: " + in + " (" + ex.getMessage() + ")");
                continue;
            }


        }
    }

    void dumpall(String s) {
        PrintStream ps = getStream(s == null ? "dump.txt" : s);
        ps.println("#### Full LC-2200 Emulator Dump (3 sections below) @ PC= " + cpu.getPc() + "####");
        dump(ps, true, true, true);
        ps.println();
    }

    void dump(PrintStream ps, boolean ph, boolean regs, boolean mem) {
        if (ph) {
            ps.println("#### LC-2200 Emulator Program History Dump at " + new Date() + "  ####");
            int i = 0;
            List<Integer> hist = cpu.getProgramHistory();
            for (int pc : hist) {
                ps.format("%10d  =>  %04d \n", i++, pc);
            }
            ps.println("#### END Program History Dump ####\n");
        }

        if (regs) {
            ps.println("#### LC-2200 Emulator Register File Dump at " + new Date() + "  ####");
            cpu.getRegisters().dump(ps);
            ps.println("#### END Register File Dump ####\n");
        }

        if (mem) {
            ps.println("#### LC2-200 Emulator Core Memory Dump at " + new Date() + "  ####");
            cpu.getMemory().dump(ps);
            ps.println("#### END Core Memory Dump ####\n");
        }

    }

    void dumpPH(String file) {
        dump(getStream(file), true, false, false);
    }

    void dumpRegs(String file) {
        dump(getStream(file), false, true, false);
    }

    void dumpMem(String file) {
        dump(getStream(file), false, false, true);
    }

    private static PrintStream getStream(String file) {
        if (file == null || (file = file.trim()).isEmpty()) {
            return System.out;
        }
        try {
            return new PrintStream(file);
        } catch (FileNotFoundException ex) {
            System.err.println("Bad File: " + file);
            return System.out;
        }
    }

    private static void helpMenu() {
        System.out.println("Commands (* commands may increment the PC): ");
        System.out.println("* [return]  - just pressing enter will clock the processor once");
        System.out.println("* run       - exe the program until it terminates, or watchdog overflows");
        System.out.println("* run X     - run the program for X number of cycles, where X >= 1");
        System.out.println("  R=X       - set REG[R] = X. Accpets $rN and rN notation");
        System.out.println("  A=X       - set MEM[A] = X. ");
        System.out.println("  in=X      - set the input variable (memory 10001) X");
        System.out.println("  pc=X      - set the PC to X");
        System.out.println("  rf [file] - dump the register file [to file]");
        System.out.println("  md [file] - dump the contents of core memory [to file]");
        System.out.println("  ph [file] - dump the program history [to file]");
        System.out.println("  dump      - dump everything to dump.txt");
        System.out.println("  reset     - restart emulator (clears everything)");
        System.out.println("  quit[d]   - quit the emulator [dump to file] ");
        System.out.println("  ?         - displays this menu ");
    }
}
