/*
 * The MIT License
 *
 * Copyright 2011 Jacob Dilles.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package edu.gmu.ece445.lc2200;

import java.io.IOException;
import java.io.PrintStream;
import java.io.Reader;

/**
 * Defines the interface to main memory, used for the emulator and simulator
 * 
 * @author Jacob Dilles
 */
public interface MainMemory {

    /**
     * Address that holds the 32 bit output (10000) for outputs to the real world
     */
    public static final int OUTPUT_ADDRESS = 10000;
        /**
     * Address that holds the 32 bit input (10001) for inputs to the real world
     */
    public static final int INPUT_ADDRESS = 10001;    
    /**
     * Address that holds the watchdog timer value.
     * Basically the watchdog is incremented every time the processor is clocked.
     * If it reaches some critical value (say, Integer.MAX_VALUE) then the 
     * CPU will halt. This helps to prevent unintentional endless loops from
     * hanging the emulator. If you are trying to intentionally loop forever,
     * just write a "0" to this address every once in a while. 
     */
    public static final int WATCHDOG_ADDRESS = 10002;
    
    /**
     * Address that holds the address of the ISR (interrupt service routine), 
     * or 0 to disable interrupts. 
     */
    public static final int ISR_PTR_ADDRESS = 10003;
    
    /**
     * Clear main memory (everything is set to zero)
     */
    void clear();

    /**
     * Get a int[n][2] where n is the number of words stored in memory. The
     * format is [[address,value],[address,value],[address,value]]
     * @return
     */
    int[][] contents();

    /**
     * Load a word from memory
     * @param address
     * @return 0(address)
     */
    int loadWord(int address);

    /**
     * Store a word in memory
     * @param address 0(address)
     * @param word to store
     * @return previous value
     */
    int storeWord(int address, int word);
    
    /**
     * Dump the contents of memory to the given printstream
     * @param out 
     */
    void dump(PrintStream out);
    
    /**
     * Load values into memory, in the format:
     * AAAAAAAA : VVVVVVVV
     * where A is the address and V is the value
     * @param in
     * @return 
     */
    int preload(Reader in) throws IOException;
}
