/*
 * The MIT License
 *
 * Copyright 2011 Jacob Dilles.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package edu.gmu.ece445.lc2200.emulator;

import edu.gmu.ece445.lc2200.AssemblyInstruction;
import edu.gmu.ece445.lc2200.LC2200;
import edu.gmu.ece445.lc2200.EncodedInstruction;
import edu.gmu.ece445.lc2200.MainMemory;
import edu.gmu.ece445.lc2200.Register;
import static edu.gmu.ece445.lc2200.Register.*;
import edu.gmu.ece445.lc2200.RegisterFile;
import edu.gmu.ece445.lc2200.Utils;
import java.util.LinkedList;
import java.util.List;

/**
 * Behavioral implementation of the LC2200 
 * @author Jacob Dilles
 */
public class LC2200Impl implements LC2200 {

    /**
     * See MainMemory.WATCHDOG_ADDRESS for info 
     */
    public static boolean WATCHDOG_ENABLED = true;
    public static int MAX_HISTORY = 512 * 1024;
    private long step = 0;
    private int pc = 0;
    private EncodedInstruction currentInstruction;
    private List<EncodedInstruction> program;
    private MainMemory memory;
    private RegisterFile registers;
    private LinkedList<Integer> programHistory = new LinkedList<Integer>();

    private long execTime = 0;
    /**
     * Create a new instance of memory and registers
     * @param program 
     */
    public LC2200Impl(List<EncodedInstruction> program) {
        this(program, new MainMemoryImpl(), new RegisterFileImpl());
    }

    /**
     * Create a new processor implementation 
     * @param program
     * @param memory
     * @param registers 
     */
    public LC2200Impl(List<EncodedInstruction> program, MainMemoryImpl memory, RegisterFileImpl registers) {
        if (program.isEmpty()) {
            throw new IllegalArgumentException("Program must have at least one instruction");
        }
        if (memory == null) {
            throw new NullPointerException("Processor requires memory");
        }
        if (registers == null) {
            throw new NullPointerException("Processor requires registers");
        }
        this.program = program;
        this.memory = memory;
        this.registers = registers;

    }

    /**
     * Clock the processor. The instruction in program[pc] is read out, then
     * the PC is incremented.
     * 
     * The instruction is then executed. This may affect memory and/or registers
     * 
     * @throws RuntimeException if the program has halted due to an error or
     *  explicit halt
     * @return true if calling clock() again will execute another instruction
     */
    @Override
    public boolean clock() throws RuntimeException {
        // Validate program counter
        if (pc < 0) {
            return false;
        }
        if (pc >= program.size()) {
            pc = -1;
            return false;
        }

        long time = System.currentTimeMillis();
        
        // get instruction and increment counter        
        setHistory(pc);
        currentInstruction = program.get(pc++);

        // execute the instruction
        int rv = execute(currentInstruction);

        // Detect unintentional endless loops
        if (WATCHDOG_ENABLED) {
            rv = memory.loadWord(MainMemory.WATCHDOG_ADDRESS) + 1;
            if (rv >= MAX_HISTORY) {
                rv = -1;
            }
            memory.storeWord(MainMemory.WATCHDOG_ADDRESS, rv);
        }

        // check for halt
        if (rv < 0) {
            pc = -1;
            throw new RuntimeException("Processor halt " + rv + (rv == -1?" (max steps exceeded)":(rv == -2 ? " (unrecognized opcode) ": "")));
        }

        // Instruction executed
        step++;
        
        execTime += System.currentTimeMillis() - time;

        // continue evaluation 
        return true;
    }

    private void setHistory(int history) {
        if (programHistory.size() > MAX_HISTORY) {
            programHistory.removeFirst();
        }
        programHistory.addLast(pc);
    }

    /**
     * Execute one instruction 
     * @param inst
     * @return 
     */
    private int execute(EncodedInstruction inst) {
        // This is legal because all 16 registers are defined
        Register rx = inst.getRegisterX(), ry = inst.getRegisterY(), rz = inst.getRegisterZ();
        // This is legal because rz is basically the lowest 4 bits of I
        int i = inst.getI();

        switch (inst.getInstruction()) {
            case add:
                return add(rx, ry, rz);
            case nand:
                return nand(rx, ry, rz);
            case addi:
                return addi(rx, ry, i);
            case lw:
                return lw(rx, ry, i);
            case sw:
                return sw(rx, ry, i);
            case beq:
                return beq(rx, ry, i);
            case jalr:
                return jalr(rx, ry);
            case halt:
                return -1;
            case sub:
                return sub(rx, ry, rz);
            case nor:
                return nor(rx, ry, rz);
            case sll:
                return sll(rx, ry, i);
            case srl:
                return srl(rx, ry, i);
            case sra:
                return sra(rx, ry, i);
            case bne:
                return bne(rx, ry, i);
            case nop:
                return 0;
            default:
                System.err.println("Unrecognized OpCode: " + inst.toHexString());
                return -2;
        }
    }

    private int add(Register rx, Register ry, Register rz) {
        int y = registers.get(ry);
        int z = registers.get(rz);
        int x = y + z;
        registers.set(rx, x);
        return 0;
    }

    private int sub(Register rx, Register ry, Register rz) {
        int y = registers.get(ry);
        int z = registers.get(rz);
        int x = y - z;
        registers.set(rx, x);
        return 0;
    }

    private int nand(Register rx, Register ry, Register rz) {
        int y = registers.get(ry);
        int z = registers.get(rz);
        int x = ~(y & z);
        registers.set(rx, x);
        return 0;
    }

    private int nor(Register rx, Register ry, Register rz) {
        int y = registers.get(ry);
        int z = registers.get(rz);
        int x = ~(y | z);
        registers.set(rx, x);
        return 0;
    }

    private int addi(Register rx, Register ry, int i) {
        int y = registers.get(ry);
        int x = y + i;
        registers.set(rx, x);
        return 0;
    }

    private int sll(Register rx, Register ry, int i) {
        int y = registers.get(ry);
        int x = y << i;
        registers.set(rx, x);
        return 0;
    }

    private int srl(Register rx, Register ry, int i) {
        int y = registers.get(ry);
        int x = y >>> i;
        registers.set(rx, x);
        return 0;
    }

    private int sra(Register rx, Register ry, int i) {
        int y = registers.get(ry);
        int x = y >> i;
        registers.set(rx, x);
        return 0;
    }

    private int lw(Register rx, Register ry, int i) {
        int y = registers.get(ry);
        int ma = y + i;
        int x = memory.loadWord(ma);
        registers.set(rx, x);
        return 0;
    }

    private int sw(Register rx, Register ry, int i) {
        int x = registers.get(rx);
        int y = registers.get(ry);
        int ma = y + i;
        memory.storeWord(ma, x);
        return 0;
    }

    private int beq(Register rx, Register ry, int i) {
        int x = registers.get(rx);
        int y = registers.get(ry);
        if (x == y) {
            pc = pc + i;
        }
        return 0;
    }

    private int bne(Register rx, Register ry, int i) {
        int x = registers.get(rx);
        int y = registers.get(ry);
        if (x != y) {
            pc = pc + i;
        }
        return 0;
    }

    private int jalr(Register rx, Register ry) {
        int x = registers.get(rx);
        registers.set(ry, pc);
        pc = x;
        return 0;
    }

    /**
     * Get the current program counter value
     */
    @Override
    public int getPc() {
        return pc;
    }

    /**
     * Set the program counter
     * @param pc 
     */
    @Override
    public void setPc(int pc) {
        this.pc = pc;
    }

    /**
     * Get the instruction which was executed by the LC-2200 the last time that
     * clock() was called
     * @return 
     */
    @Override
    public EncodedInstruction getCurrentInstruction() {
        return currentInstruction;
    }

    /**
     * Get a copy of the current register contents. The array has 16 elements.
     * @return 
     */
    @Override
    public int[] getRegisterContents() {
        return registers.contents();
    }

    /**
     * Get a int[n][2] where n is the number of words stored in memory. The 
     * format is [[address,value],[address,value],[address,value]]
     * @return 
     */
    @Override
    public int[][] getMemoryContents() {
        return memory.contents();
    }

    /**
     * Get the main memory used by the program. The object is not copied so 
     * changing the returned object will affect processor opperation 
     * @return 
     */
    @Override
    public MainMemory getMemory() {
        return memory;
    }

    /**
     * Get the registers used by the program. The object is not copied so
     * changes to the returned object will be reflected by processor operation 
     * @return 
     */
    @Override
    public RegisterFile getRegisters() {
        return registers;
    }

    @Override
    public String toString() {
        if (pc < 0) {
            double seconds = ((double)execTime)/(1000.0);
            double mhz = (1.0/(seconds / ((double)step)));
            int instRate = (int)mhz;
            mhz /= (1024.0*1024.0);
            
            return String.format("@%010d:  HALT - %f mS total (%f MHz or %d ops/sec)", execTime, seconds, mhz, instRate);
        }
        return String.format("@%010d:  @%04d  WAIT ", step, pc);
    }

    @Override
    public List<EncodedInstruction> getProgram() {
        return program;
    }

    @Override
    public List<Integer> getProgramHistory() {
        return programHistory;
    }

    public long getStep() {
        return step;
    }

    @Override
    public String report() {
        if (currentInstruction == null || pc < 0) {
            return toString();
        }
        AssemblyInstruction i2;
        if (currentInstruction instanceof AssemblyInstruction) {
            i2 = (AssemblyInstruction) currentInstruction;
        } else {
            i2 = new AssemblyInstruction(currentInstruction.getWord());
        }
        return String.format("%10d:  @%04d  %-5s %20s  OUT=%s", step, pc, i2.getName(), i2.getArgsString(), memory.loadWord(MainMemory.OUTPUT_ADDRESS))
                +"\n"
                + String.format("sp=%s fp=%s ra=%s a0=%s a1=%s a2=%s", rh(sp), rh(fp), rh(ra), rh(a0), rh(a1), rh(a2)) +"\n"
                + String.format("t0=%s t1=%s t2=%s s0=%s s1=%s s2=%s", rh(t0), rh(t1), rh(t2), rh(s0), rh(s1), rh(s2));
// step    pc    curr command  cmd args 
// sp fp ra a0 a1 a2 
// t0 t1 t2 s0 s1 s2

    }

    private String rh(Register r) {
        return Utils.toHex(registers.get(r));
    }
}
