/*
 * The MIT License
 *
 * Copyright 2011 Jacob Dilles.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package edu.gmu.ece445.lc2200.emulator;

import edu.gmu.ece445.lc2200.MainMemory;
import edu.gmu.ece445.lc2200.Utils;
import java.io.IOException;
import java.io.PrintStream;
import java.io.Reader;
import java.util.Map.Entry;
import java.util.Properties;
import java.util.TreeMap;

/**
 * Behavioral emulation of the LC-2200 main memory 
 * 
 * Storing memory in an array would be impractical due to the 32-bit address bus.
 * Instead, it is sparsely implemented using a TreeMap
 * 
 * @author Jacob Dilles
 */
public class MainMemoryImpl implements MainMemory {

    /**
     * Key = address, Value = contents 
     */
    private TreeMap<Integer, Integer> memory = new TreeMap<Integer, Integer>();

    public MainMemoryImpl() {
    }

    /**
     * Store a word in memory
     * @param address 0(address)
     * @param word to store
     * @return previous value
     */
    @Override
    public int storeWord(int address, int word) {
        Integer rv = memory.put(address, word);
        return  rv == null ? 0 : rv;
    }

    /**
     * Load a word from memory
     * @param address
     * @return 0(address)
     */
    @Override
    public int loadWord(int address) {
        Integer rv = memory.get(address);
        return rv == null ? 0 : rv;
    }

    /**
     * Clear main memory (everything is set to zero)
     */
    @Override
    public void clear() {
        memory.clear();
    }

    /**
     * Get a int[n][2] where n is the number of words stored in memory. The 
     * format is [[address,value],[address,value],[address,value]]
     * @return 
     */
    @Override
    public int[][] contents() {
        int[][] rv = new int[memory.size()][2];
        int i = 0;
        for (Entry<Integer, Integer> e : memory.entrySet()) {
            rv[i][0] = e.getKey();
            rv[i][1] = e.getValue();
            i++;
        }
        return rv;
    }

    @Override
    public void dump(PrintStream out) {
        for (Entry<Integer, Integer> e : memory.entrySet()) {
            out.printf("%8s=%8s\n", Utils.toHex(e.getKey()), Utils.toHex(e.getValue()));
        }
    }

    @Override
    public int preload(Reader in) throws IOException {
        clear();
        Properties props = new Properties();
        props.load(in);
        for(Entry e : props.entrySet()){
            storeWord(Integer.parseInt(e.getKey().toString(),16), Integer.parseInt(e.getValue().toString(), 16));
        }
        return memory.size();
    }
    
    
}
