/*
 * The MIT License
 *
 * Copyright 2011 Jacob Dilles.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package edu.gmu.ece445.lc2200.emulator;

import edu.gmu.ece445.lc2200.RegisterFile;
import edu.gmu.ece445.lc2200.Register;
import edu.gmu.ece445.lc2200.Register;
import edu.gmu.ece445.lc2200.Utils;
import edu.gmu.ece445.lc2200.Utils;
import java.io.PrintStream;
import java.util.EnumMap;
import java.util.Map.Entry;

/**
 * Behavioral emulation of the LC-2200 Register File
 * @author Jacob Dilles
 */
public class RegisterFileImpl implements RegisterFile {

    private int[] ram = new int[Register.REGISTER_COUNT];

    /**
     * Creates an empty register file
     */
    public RegisterFileImpl() {
        clear();
    }

    /**
     * Get a register value
     * @param reg
     * @return 
     */
    @Override
    public int get(Register reg) {
        return ram[reg.getNumber()];
    }

    /**
     * Set a register value
     * @param register destination
     * @param value new value 
     * @return the old value 
     */
    @Override
    public int set(Register register, int value) {
        int rv = ram[register.getNumber()];
        // zero is R/O by hardware
        if (register != Register.zero) {
            ram[register.getNumber()] = value;
        }
        return rv;
    }

    /**
     * Set all registers to 0
     */
    @Override
    public void clear() {
        for(int i = 0; i < ram.length; i++){
            ram[i] =0;
        }
    }

    /**
     * Dump register contents 
     * @param out 
     */
    @Override
    public void dump(PrintStream out) {
        for (Register reg : Register.values()) {
            out.printf("    %6s : %8s \n", reg.toString(), Utils.toHex(get(reg)));
        }
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder(((2+8+4)*Register.REGISTER_COUNT) + 16);
        sb.append("RegisterFile[");
        for (Register reg : Register.values()) {
            sb.append(reg.toString());
            sb.append("=");
            sb.append(Utils.toHex(get(reg)));
            sb.append(", ");
        }
        sb.setLength(sb.length()-2);
        sb.append("]");
        return sb.toString();
    }
    
    
    /**
     * Get the contents of the register file. The array has 16 elements
     * @return 
     */
    @Override
    public int[] contents() {
        int[] contents = new int[16];
        for (int i = 0; i < 16; i++) {
            contents[i] = get(Register.valueOfNumber(i));
        }
        return contents;
    }
}
