/*
 * The MIT License
 *
 * Copyright 2011 Jacob Dilles <jdilles@gmu.edu>.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

import edu.gmu.ece445.lc2200.AssemblyInstruction;
import edu.gmu.ece445.lc2200.AssemblyProgram;
import edu.gmu.ece445.lc2200.Emulator;
import edu.gmu.ece445.lc2200.EncodedInstruction;
import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.io.OutputStream;
import java.io.PrintStream;
import java.util.Date;
import java.util.List;

/**
 * Main program. Just a quick wrapper for the edu.gmu.ece445.lc2200 classes
 * to allow command-line interfacing  
 * 
 * @author Jacob Dilles
 */
public class lcasm {

    public static boolean VERBOSE = true;
    public static boolean ENCODE_IN_HEX = true;
    public static boolean ENCODE_IN_AEB = false;
    public static boolean ENCODE_IN_RAW = false;
    public static boolean ENCODE_LINEBREAKS = true;
    public static boolean ANNOTATE_ASCII_OUTPUT = true;
    public static boolean TEST_ASSEMBLE_ONLY = false;
    public static PrintStream ASM_OUTPUT = System.out;
    public static PrintStream ASM_ERRORS = System.err;
    public static int RETURN_VAL = 0;

    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) throws IOException {
        if (args.length < 2 || args[0].isEmpty()) {
            banner();
            System.exit(1);
        }
        System.out.println("George Mason University LC-2200 Assembler V0.3 (2011)                          ");

        File infile = new File(args[1]);
        if (!infile.isFile()) {
            System.err.println("IO Error, file missing: " + infile);
        }
        InputStream infileStream = new BufferedInputStream(new FileInputStream(infile));

        String c = args[0];

        if (c.startsWith("e")) {

            AssemblyProgram program = lcasm.assemble(infileStream);
            Emulator em = new Emulator(program);

            // Write the program to output 
            if (RETURN_VAL == 0) {
                em.emulate(c);
            } 



        } else // Assemble option 
        if (c.startsWith("a")) {

            // Parse flags
            parseAssembleFlags(c);


            //Figure out the output 
            String outputFileName = "output";
            if (TEST_ASSEMBLE_ONLY) {
                ASM_OUTPUT = System.out;
                outputFileName = null;
            } else if (args.length >= 3) {
                outputFileName = new File(args[2]).getAbsolutePath();
                ASM_OUTPUT = new PrintStream(outputFileName + ".asm");
            } else {
                outputFileName = infile.getAbsolutePath();
                int lio = outputFileName.lastIndexOf(".");
                if (lio > 0) {
                    outputFileName = outputFileName.substring(0, lio);
                }
            }


            // Assemble the program 
            AssemblyProgram prog = assemble(infileStream);
            if (RETURN_VAL != 0) {
                System.exit(RETURN_VAL);
            }

            if (TEST_ASSEMBLE_ONLY) {
                ASM_OUTPUT.println();
            }

            // Write the program to output 
            if (RETURN_VAL == 0) {
                encode(prog, outputFileName);
            }


        } else {
            System.err.println("Unknown options: " + c);
            RETURN_VAL = 1;
        }

        System.exit(RETURN_VAL);
    }

    /**
     * Writes the program to the various output files 
     * @param format
     * @param in
     * @param out
     * @throws IOException 
     */
    private static void encode(AssemblyProgram prog, String out) throws IOException {
        List<EncodedInstruction> insts = prog.getEncodedProgram();

        PrintStream ps;

        if (ENCODE_IN_HEX) {
            ps = out == null ? ASM_OUTPUT : new PrintStream(out + "_hex.txt");
            if (ANNOTATE_ASCII_OUTPUT) {
                ps.println("### LC-2200 Assembler hex output - " + new Date() + " ###");
            }
            for (EncodedInstruction inst : insts) {
                ps.print(inst.toHexString());
                if (ANNOTATE_ASCII_OUTPUT) {
                    ps.print("    #");
                    ps.print(inst.toString());
                }
                if (ENCODE_LINEBREAKS) {
                    ps.print("\n");
                }
            }

            //          ### LC-2200 Assembler output - Thu Sep 22 12:15:29 EDT 2011 ###
            if (ANNOTATE_ASCII_OUTPUT) {
                ps.println("###          End LC-2200 Assembler output                   ###");
            }
            if (out != null) {
                ps.close();
            } else {
                ps.print("\n");
            }
        }

        if (ENCODE_IN_AEB) {
            ps = out == null ? ASM_OUTPUT : new PrintStream(out + "_binary.txt");
            if (ANNOTATE_ASCII_OUTPUT) {
                ps.println("### LC-2200 Assembler binary output - " + new Date() + " ###");
            }
            for (EncodedInstruction inst : insts) {
                ps.print(inst.toBinaryString());
                if (ENCODE_LINEBREAKS) {
                    ps.print("\n");
                }
            }

            //          ### LC-2200 Assembler output - Thu Sep 22 12:15:29 EDT 2011 ###
            if (ANNOTATE_ASCII_OUTPUT) {
                ps.println("###             End LC-2200 Assembler output                   ###");
            }
            if (out != null) {
                ps.close();
            } else {
                ps.print("\n");
            }
        }

        if (ENCODE_IN_RAW && out != null) {
            OutputStream os = new FileOutputStream(out + ".dat");
            for (EncodedInstruction inst : insts) {
                os.write(inst.getBytes());
            }
            os.close();
        }
    }

    /**
     * Assemble the program
     * @param in
     * @param echo
     * @param errors
     * @return
     * @throws IOException 
     */
    public static AssemblyProgram assemble(InputStream in) throws IOException {

        // Parse the program
        LineNumberReader lnr = new LineNumberReader(new InputStreamReader(in));
        String line;
        StringBuilder eb = new StringBuilder(256);
        long time = System.currentTimeMillis();
        AssemblyProgram prog = new AssemblyProgram();
        String assmInstParsed;
        int errorCount = 0;

        while ((line = lnr.readLine()) != null) {
            line = line.trim();
            if (line.isEmpty() || line.startsWith(";") || line.startsWith("#") || line.startsWith("-")) {
                ASM_ERRORS.println(";### Ignoring line " + lnr.getLineNumber() + "  (empty)");
                continue;
            }
            eb.setLength(0);

            try {
                AssemblyInstruction inst = AssemblyProgram.parseInstruction(line);
                prog.addInstruction(inst);
                assmInstParsed = inst.toString();
                eb.append(assmInstParsed);
            } catch (Exception ex) {
                ASM_ERRORS.println("Line " + lnr.getLineNumber() + ": " + ex.getMessage());
                ex.printStackTrace(ASM_ERRORS);
                eb.append(";#### ");
                eb.append(line);
                eb.append("  ##### ERROR:  ");
                eb.append(ex.getMessage());
                errorCount++;
            }
            if (VERBOSE) {
                ASM_OUTPUT.println(eb);
            }

        }

        // Parse stats
        time = System.currentTimeMillis() - time;
        if (VERBOSE) {
            ASM_OUTPUT.println(";### Parsed " + prog.size() + " instructions in " + time + " mS    [" + errorCount + " error" + (errorCount != 1 ? "s" : "") + "]");
        }


        // Link the prgoram 
        time = System.currentTimeMillis();
        int links = prog.assembleProgram(VERBOSE ? ASM_OUTPUT : null);

        // Link stats 
        time = System.currentTimeMillis() - time;
        if (VERBOSE) {
            ASM_OUTPUT.println(";### Linked " + links + " instruction" + (links != 1 ? "s" : "") + " in " + time + " mS     [0 errors]");
        }


        RETURN_VAL = errorCount;

        return prog;

    }

    /**
     * Prints the help screen
     */
    private static void banner() {
        System.out.println("George Mason University LC-2200 Assembler V0.3 (2011)                          ");
        System.out.println("Usage:                                                                         ");
        System.out.println("  java -jar lcasm.jar a[bBxXrRs] input.asm [output]                            ");
        System.out.println("                                                                               ");
        System.out.println("         a(ssemble) takes the following flags flags:                           ");
        System.out.println("                 q   - quiet, suppress assembler output                        ");
        System.out.println("                 s   - simulate (parse, validate, do not write output)         ");
        System.out.println("               b[B]  - do [not] write ascii-encoded binary file  (.b)          ");
        System.out.println("               x[X]  - do [not] write ascii-encoded hex file     (.hex)        ");
        System.out.println("               t[T]  - do [not] annoTate ascii output (implies l)              ");
        System.out.println("               n[N]  - do [not] add per-instruction linebreaks to ascii output ");
        System.out.println("               r[R]  - do [not] write raw binary file            (.dat)        ");
        System.out.println("                                                                               ");
        System.out.println("  java -jar lcasm.jar e input.asm ");
        System.out.println("");
        System.out.println("         e(mulate) lauches the interactive emulator ");
        System.out.println("                 q   - quiet, supress per-clock output");
        System.out.println("                 p   - specifiy an initial program counter value (prompt)      ");
        System.out.println("                 m   - activate the 'load core' prompt                         ");
        System.out.println("                 r   - just run the program, and dump when complete            ");
        System.out.println("\n\n");
        System.out.println("  Send bug reports to: Jacob Dilles <jdilles@gmu.edu> ");
    }

    private static void parseAssembleFlags(String c) {
        loop:
        for (int i = 1; i < c.length(); i++) {
            switch (c.charAt(i)) {
                case 'b':
                    ENCODE_IN_AEB = true;
                    continue loop;
                case 'B':
                    ENCODE_IN_AEB = false;
                    continue loop;
                case 'x':
                    ENCODE_IN_HEX = true;
                    continue loop;
                case 'X':
                    ENCODE_IN_HEX = false;
                    continue loop;
                case 'r':
                    ENCODE_IN_RAW = true;
                    continue loop;
                case 'R':
                    ENCODE_IN_RAW = false;
                    continue loop;
                case 's':
                    TEST_ASSEMBLE_ONLY = true;
                    continue loop;
                case 'a':
                case 'A':
                    ENCODE_IN_RAW = true;
                    ENCODE_IN_AEB = true;
                    ENCODE_IN_HEX = true;
                    continue loop;
                case 'n':
                    ENCODE_LINEBREAKS = true;
                    continue loop;
                case 'N':
                    ENCODE_LINEBREAKS = false;
                    continue loop;
                case 't':
                    ANNOTATE_ASCII_OUTPUT = true;
                    continue loop;
                case 'T':
                    ANNOTATE_ASCII_OUTPUT = false;
                    continue loop;
                case 'q':
                    VERBOSE = false;
                    continue loop;
                default:
                    System.err.println("Unrecognized (a)ssemble flag: " + c.charAt(i));
            }
        }
        ENCODE_LINEBREAKS = ENCODE_LINEBREAKS || ANNOTATE_ASCII_OUTPUT;
    }
}
